/***************************************************************************
                          cclient.cpp  -  description
                             -------------------
    begin                : Sun Sep 30 2001
    copyright            : (C) 2001-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>

#ifndef WIN32
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <dclib/dcos.h>
#include <dclib/clist.h>
#include <dclib/dcobject.h>
#include <dclib/cquerymanager.h>
#include <dclib/cdownloadmanager.h>
#include <dclib/cservermanager.h>
#include <dclib/cencrypt.h>
#include <dclib/cmessagehandler.h>
#include <dclib/cstringlist.h>
#include <dclib/cconfig.h>
#include <dclib/chubsearch.h>

#include "cclient.h"

CClient::CClient()
{
	m_pCallback = 0;

	m_ecMode = ecmNone;

	m_sBuffer          = "";
	m_sNick            = "";
	m_sEMail           = "";
	m_sConnectionType  = "";
	m_sComment         = "";
	m_sHubName         = "";
	m_sShareSize       = "";
	m_sVersion         = "";
	m_bHandshake       = TRUE;
	m_bHandleUserList  = TRUE;
	m_bHandleSearch    = TRUE;
	m_bHandleMyinfo    = TRUE;
	m_bHandleForceMove = TRUE;
	m_bHandleTransfer  = TRUE;
	m_bUpdateMyinfo    = FALSE;
	m_nHubShareSize    = 0;
	m_nNickListHandler = 0;
	m_nOpListHandler   = 0;
	m_eAwayMode        = euamNONE;

	// set reconnect parameter
	m_eReconnectState      = ersNONE;
	m_nReconnectCount      = 0;
	m_timeReconnectTimeout = 0;
	m_timeMyinfoTimeout    = 0;

	m_pUserList = new CThreadStringList();
}

CClient::~CClient()
{
	SetCallBackFunction(0);

	if ( m_pUserList )
	{
		delete m_pUserList;
		m_pUserList = 0;
	}
}

/** */
void CClient::ConnectionState( eConnectionState state )
{
	int err = -1;

	CMessageConnectionState *Object = new CMessageConnectionState();

	if (!Object)
	{
		return;
	}

	if ( (state == estCONNECTED) ||
	     (state == estDISCONNECTED) )
	{
		// reset all values
		m_sBuffer           = "";
		m_bHandshake        = TRUE;
		m_nHubShareSize     = 0;
		m_timeMyinfoTimeout = time(0);
		m_nNickListHandler  = 0;
		m_nOpListHandler    = 0;

		if ( m_pUserList )
		{
			m_pUserList->Lock();
			m_pUserList->Clear();
			m_pUserList->UnLock();
		}
#ifdef HAVE_SSL
		m_pMessageSSL.Init();
#endif
		// update reconnect state
		if ( state == estCONNECTED )
		{
			UpdateReconnect( ersNONE );
		}
		else
		{
			if ( m_eReconnectState == ersFORCEMOVE )
			{
				UpdateReconnect( ersNONE );
			}
			else
			{
				UpdateReconnect( ersENABLED );
			}
		}
	}

	Object->m_eType  = DC_MESSAGE_CONNECTION_STATE;
	Object->eState   = state;
	Object->sMessage = GetSocketError();

	if ( m_pCallback != 0 )
	{
		err = m_pCallback->notify( this, Object );
	}
	else
	{
		err = DC_CallBack( Object );
	}

	// callback failed
	if ( err == -1 )
	{
		delete Object;
	}

	// update myinfo
	if ( m_bHandleMyinfo )
	{
		CServerManager::Instance()->SendMyInfoToConnectedServers();
	}
}

/** */
void CClient::DataAvailable( const char * buffer, int len )
{
	CString s;
	int i,i_old;

	if ( len <= 0 )
	{
		return;
	}

	s.Set(buffer,len);

	// add traffic control
	CSocket::m_Traffic.AddTraffic(ettCONTROLRX,len);

	i = i_old = 0;

	s = m_sBuffer + s;

	// search '|'
	while((i=s.Find('|',i))>=0)
	{
		i++;
		i_old = i;
	}

	if ( i_old > 0 )
	{
		m_sBuffer = s.Mid(0,i_old);
		HandleMessage( m_sBuffer );
	}

	if (i_old==s.Length())
	{
		m_sBuffer = "";
	}
	else
	{
		m_sBuffer = s.Mid( i_old, s.Length()-i_old );
	}
}

/** */
void CClient::DataTimeout()
{
	printf("CClient: data timeout ... check connection\n");
	
	SendString("|");
}

/** */
void CClient::Notify()
{
	// check away mode
	if ( (m_bHandshake == FALSE) && ((time(0)-m_timeMyinfoTimeout) >= 30) )
	{
		if ( CConfig::Instance()->GetAwayMode() != m_eAwayMode)
		{
			m_eAwayMode = CConfig::Instance()->GetAwayMode();
			m_bUpdateMyinfo = TRUE;
		}
	}

	if ( (m_bHandshake == FALSE) && (m_bUpdateMyinfo == TRUE) && ((time(0)-m_timeMyinfoTimeout) >= 30) )
	{
		m_bUpdateMyinfo     = FALSE;
		m_timeMyinfoTimeout = time(0);

		SendMyInfo( m_sNick, m_sComment, m_sConnectionType, m_eAwayMode, m_sEMail, m_sShareSize );
	}

	if ( m_eReconnectState == ersENABLED )
	{
		if ( m_nReconnectCount >= CConfig::Instance()->GetReconnectCount() )
		{
			UpdateReconnect( ersNONE, 0 );
		}
		else
		{
			if ( GetConnectionState() != estNONE )
			{
				printf("warning, wrong reconnect state, you are connected !\n");
			}

			// init timer
			if ( m_timeReconnectTimeout == 0 )
			{
				m_timeReconnectTimeout = time(0);
			}

			if ( (time(0)-m_timeReconnectTimeout) >= CConfig::Instance()->GetReconnectTimeout() )
			{
				UpdateReconnect( ersNONE );

				if ( CConfig::Instance()->GetReconnectCount() != 9999 )
				{
					m_nReconnectCount++;
				}

				Connect();
			}
		}
	}
}

/** */
void CClient::UpdateReconnect( eReconnectState state, int count )
{
	// reconnect disabled
	if ( CConfig::Instance()->GetReconnectCount() == 0 )
	{
		m_eReconnectState = ersNONE;
		return;
	}

	// don't change state if reconnect disables
	if ( (m_eReconnectState == ersDISABLED) && (state != ersNONE) )
	{
		return;
	}

	// update state
	m_eReconnectState      = state;
	// reset timeout
	m_timeReconnectTimeout = 0;
	// update counter
	if ( count != -1 )
	{
		m_nReconnectCount = count;
	}
}

/** */
int CClient::HandleMessage( const CString message )
{
	int err;
	int pointer;
	eDCMessage type;
	CObject * Object;
	CMessageHandler MessageHandler;
	CString sAnswer;
	CEncrypt Encrypt;

	if ( message == "" )
	{
		return 0;
	}

	pointer = 0;

	while( (type=MessageHandler.Parse(&message,pointer,Object=0)) != DC_MESSAGE_PARSE_ERROR )
	{
		if ( !Object )
		{
			continue;
		}

		switch (type)
		{
			case DC_MESSAGE_LOCK:
			{
				CMessageLock * msg = (CMessageLock*)Object;

				// only handle the first lock in handshake mode
				if ( m_bHandshake == TRUE )
				{
					Encrypt.Encrypt(msg->sData,sAnswer);

					SendKey( sAnswer );

					if ( msg->bExtProtocol == TRUE )
					{
						// quicklist ??? http://forum.dcstats.net/showthread.php?s=&threadid=802
					}

					SendValidateNick( m_sNick );
				}

				break;
			}

			case DC_MESSAGE_LOGEDIN:
			{
				// update myinfo
				if ( m_bHandleMyinfo )
					CServerManager::Instance()->SendMyInfoToConnectedServers();

				break;
			}

			case DC_MESSAGE_HUBNAME:
			{
				CMessageHubName * msg = (CMessageHubName*)Object;

				// set the old hubname in this message
				if ( GetHubName() != msg->sHubName )
					msg->sOldHubName = GetHubName();
				SetHubName(msg->sHubName);

				break;
			}

			case DC_MESSAGE_HELLO:
			{
				CMessageHello * msg = (CMessageHello*)Object;

				if ( (msg->sNick == m_sNick) && (m_bHandshake == TRUE) )
				{
					SendVersion( m_sVersion );

					// get nicklist
					if ( m_bHandleUserList == TRUE )
						RequestNickList();

					SendMyInfo( m_sNick, m_sComment, m_sConnectionType, m_eAwayMode, m_sEMail, m_sShareSize );

					m_bHandshake = FALSE;

					// update myinfo
					if ( m_bHandleMyinfo )
						CServerManager::Instance()->SendMyInfoToConnectedServers();
				}
				else if ( m_bHandleUserList == TRUE )
				{
					AppendUser(msg->sNick);
#ifdef HAVE_SSL
					m_pMessageSSL.JoinHub(this,msg->sNick);
#endif
				}

				break;
			}

			case DC_MESSAGE_MYINFO:
			{
				CMessageMyInfo * msg = (CMessageMyInfo*)Object;

				if( m_bHandleUserList == TRUE )
				{
					if ( UpdateUser(msg) == FALSE )
					{
						delete Object;
						Object = 0;
					}
				}

				break;
			}

			case DC_MESSAGE_QUIT:
			{
				CMessageQuit * msg = (CMessageQuit*)Object;

				if ( m_bHandleUserList == TRUE )
				{
					RemoveUser(msg->sNick);
#ifdef HAVE_SSL
					m_pMessageSSL.LeaveHub(this,msg->sNick);
#endif
				}

				break;
			}

			case DC_MESSAGE_NICKLIST:
			{
				CMessageNickList * msg = (CMessageNickList*)Object;

				m_nHubShareSize = 0;

				if( (m_pUserList) && (m_bHandleUserList == TRUE))
				{
					// 0: update nicklist 1: requested nicklist
					if ( m_nNickListHandler == 1 )
					{
						m_pUserList->Lock();
						m_pUserList->Clear();
						m_pUserList->UnLock();

						m_nNickListHandler = 0;
					}

					InitUserList(msg);
				}

				break;
			}

			case DC_MESSAGE_OPLIST:
			{
				CMessageNickList * msg = (CMessageNickList*)Object;

				// operators allready in the userlist
				if( (m_pUserList) && (m_bHandleUserList == TRUE))
				{
					// init operator list
					InitOperatorList(msg);
				}
				break;
			}

			case DC_MESSAGE_SEARCH:
			{
				bool search = TRUE;
				CMessageSearch * msg = (CMessageSearch*)Object;

				// check for loopback search
				if ( msg->bLocal )
				{
					if ( msg->sSource == GetNick() )
					{
						search = FALSE;
					}
				}
				else
				{
					// TODO: disable global search, compare with own ip is slow (dns) !
				}

				if ( (search == TRUE) && (CQueryManager::Instance() != 0) && (m_bHandleSearch) )
				{
					CQueryManager::Instance()->SearchQuery( GetHubName(), GetHost(), GetNick(), msg );
				}

				break;
			}

			case DC_MESSAGE_SEARCHRESULT:
			{
				if( (m_pUserList) && (m_bHandleUserList == TRUE) )
				{
					UpdateUserSlots( (CMessageSearchResult *) Object );
				}

				if ( CHubSearch::Instance() )
					if ( CHubSearch::Instance()->HandleMessage( (CMessageSearchResult *) Object ) == TRUE )
						Object = 0;
				break;
			}

			case DC_MESSAGE_CONNECTTOME:
			{
				CMessageConnectToMe * msg = (CMessageConnectToMe*) Object;

				if ( m_bHandleTransfer == TRUE )
				{
					CDownloadManager::Instance()->DLM_AddTransferRequest( msg->sHost, msg->iPort, GetHubName(), GetHost() );
				}

				break;
			}

			case DC_MESSAGE_REVCONNECTTOME:
			{
				CMessageRevConnectToMe * msg = (CMessageRevConnectToMe*) Object;

				if ( (GetMode() == ecmActive) && (m_bHandleTransfer == TRUE) )
				{
					CString s = CConfig::Instance()->GetTCPHostString();

					if ( s != "" )
					{
						if ( CDownloadManager::Instance()->DLM_AddTransferRequest( msg->sRemoteNick, "", GetHubName(), GetHost() ) == TRUE )
						{
							SendConnectToMe(msg->sRemoteNick,s);
						}
					}
				}

				break;
			}

			case DC_MESSAGE_FORCEMOVE:
			{
				CMessageForceMove * msg = (CMessageForceMove*) Object;

				if ( (CConfig::Instance()->GetForceMoveEnabled() == TRUE) && (m_bHandleForceMove == TRUE) )
				{
					SetHubName(msg->sHost);

					UpdateReconnect( ersFORCEMOVE, 0 );

					Connect(msg->sHost,msg->iPort);
				}

				break;
			}

			case DC_MESSAGE_PRIVATECHAT:
			{
				CMessagePrivateChat * msg = (CMessagePrivateChat*) Object;

				if ( (CConfig::Instance()->GetChatRecvOfflineMessages() == FALSE) && (IsUserOnline(msg->sFrom) == FALSE) )
				{
					delete Object;
					Object = 0;
				}
#ifdef HAVE_SSL
				m_pMessageSSL.PrivateChat(this,msg);
#endif
				break;
			}

			case DC_MESSAGE_CAPABILITIES:
			{
				m_MessageCapabilities = *((CMessageCapabilities*) Object);
				break;
			}

			default:
			{
				break;
			}
		}

		if (Object)
		{
			((CDCMessage*)Object)->m_eType = type;

			if ( m_pCallback != 0 )
			{
				err = m_pCallback->notify( this, Object );
			}
			else
			{
				err = DC_CallBack( Object );
			}

			if ( err == -1 )
			{
				delete Object;
			}
		}
	}

	return 0;
}

/** */
void CClient::InitUserList( CMessageNickList * nicklist )
{
	CString *nick;

	if (!nicklist)
		return;
	if (!m_pUserList)
		return;

	for(nick=0;(nick=nicklist->NickList.Next(nick))!=0;)
	{
		AppendUser(*nick);
	}
}

/** */
void CClient::InitOperatorList( CMessageNickList * nicklist )
{
	CString *nick;
	CMessageMyInfo *myinfo=0;

	if (!m_pUserList)
		return;

	// set in the nicklist the operator status
	for(nick=0;(nick=nicklist->NickList.Next(nick))!=0;)
	{
		m_pUserList->Lock();
		if ( m_pUserList->Get( *nick, (CObject*&)myinfo) == 0 )
		{
			myinfo->bOperator = TRUE;
		}
		m_pUserList->UnLock();
	}
}

/** */
void CClient::AppendUser( CString nick, bool getinfo )
{
	CString s;
	CMessageMyInfo *myinfo=0;

	if (!m_pUserList)
		return;

	m_pUserList->Lock();

	if ( m_pUserList->Get( nick, (CObject*&)myinfo ) != 0 )
	{
		myinfo = new CMessageMyInfo();
		myinfo->sNick     = nick;
		myinfo->lShared   = 0;
		myinfo->bOperator = FALSE;
		myinfo->eAwayMode = euamNORMAL;

		m_pUserList->Add(nick,myinfo);

		if ( (!m_bHandshake) && (getinfo) )
			SendGetInfo(nick,GetNick());
	}

	m_pUserList->UnLock();
}

/** */
void CClient::RemoveUser( CString nick )
{
	CMessageMyInfo *myinfo=0;

	if (!m_pUserList)
		return;

	m_pUserList->Lock();

	if ( m_pUserList->Get( nick, (CObject*&)myinfo ) == 0 )
	{
		if ( myinfo->lShared > m_nHubShareSize )
			m_nHubShareSize = 0;
		else
			m_nHubShareSize -= myinfo->lShared;

		m_pUserList->Del(nick);
	}

	m_pUserList->UnLock();
}

/** */
bool CClient::UpdateUser( CMessageMyInfo * pMyInfo )
{
	bool res = FALSE;
	CMessageMyInfo *myinfo=0;

	if (!pMyInfo)
		return FALSE;
	if (!m_pUserList)
		return FALSE;

	m_pUserList->Lock();

	if ( m_pUserList->Get( pMyInfo->sNick, (CObject*&)myinfo ) == 0 )
	{
		if ( myinfo->lShared >= m_nHubShareSize )
			m_nHubShareSize = 0;
		else
			m_nHubShareSize -= myinfo->lShared;
	}
	else
	{
		myinfo = new CMessageMyInfo();
		myinfo->bOperator = FALSE;
		m_pUserList->Add(pMyInfo->sNick,myinfo);

		res = TRUE;
	}

	// don't send old data
	if ( myinfo->eAwayMode != pMyInfo->eAwayMode )
		res = TRUE;
	else if ( myinfo->lShared != pMyInfo->lShared )
		res = TRUE;
	else if ( myinfo->sComment != pMyInfo->sComment )
		res = TRUE;
	else if ( myinfo->sVerComment != pMyInfo->sVerComment )
		res = TRUE;
	else if ( myinfo->sEMail != pMyInfo->sEMail )
		res = TRUE;
	else if ( myinfo->sNick != pMyInfo->sNick )
		res = TRUE;
	else if ( myinfo->sSpeed != pMyInfo->sSpeed )
		res = TRUE;
	else if ( myinfo->sUnknown != pMyInfo->sUnknown )
		res = TRUE;
	else if ( pMyInfo->bOperator != myinfo->bOperator )
		res = TRUE;

	if ( res == TRUE )
	{
		myinfo->eAwayMode  = pMyInfo->eAwayMode;
		myinfo->lShared    = pMyInfo->lShared;
		myinfo->sComment   = pMyInfo->sComment;
		myinfo->sEMail     = pMyInfo->sEMail;
		myinfo->sNick      = pMyInfo->sNick;
		myinfo->sSpeed     = pMyInfo->sSpeed;
		myinfo->sUnknown   = pMyInfo->sUnknown;
		pMyInfo->bOperator = myinfo->bOperator;
	}

	m_nHubShareSize += myinfo->lShared;

	m_pUserList->UnLock();

	return res;
}

/** */
void CClient::UpdateUserSlots( CMessageSearchResult * pSearchResult )
{
	int err;
	CMessageMyInfo *myinfo=0,*mi;

	if (!pSearchResult)
		return;
	if (!m_pUserList)
		return;

	m_pUserList->Lock();

	if ( m_pUserList->Get( pSearchResult->sNick, (CObject*&)myinfo ) == 0 )
	{
		if ( (myinfo->m_nFreeSlot != pSearchResult->iFreeSlot) ||
		     (myinfo->m_nMaxSlot != pSearchResult->iMaxSlot) )
		{
			myinfo->m_nFreeSlot = pSearchResult->iFreeSlot;
			myinfo->m_nMaxSlot  = pSearchResult->iMaxSlot;

			mi = new CMessageMyInfo();
			*mi = *myinfo;

			if ( m_pCallback != 0 )
			{
				err = m_pCallback->notify( this, mi );
			}
			else
			{
				err = DC_CallBack( mi );
			}

			if ( err == -1 )
			{
				delete mi;
			}
		}
	}

	m_pUserList->UnLock();
}

/** */
bool CClient::IsUserOnline( CString nick )
{
	bool res = FALSE;
	CMessageMyInfo *myinfo=0;

	if (!m_pUserList)
		return res;

	m_pUserList->Lock();
	res = ( m_pUserList->Get( nick, (CObject*&)myinfo ) == 0 );
	m_pUserList->UnLock();

	return res;
}

/** */
bool CClient::IsAdmin() const
{
	return IsAdmin( GetNick() );
}

/** */
bool CClient::IsAdmin( CString nick ) const
{
	bool res = FALSE;
	CMessageMyInfo *myinfo=0;

	if (!m_pUserList)
		return res;

	m_pUserList->Lock();
	if ( m_pUserList->Get( nick, (CObject*&)myinfo ) == 0 )
		res = myinfo->bOperator;
	m_pUserList->UnLock();

	return res;
}

/** send private message */
int CClient::SendPrivateMessage( CString sNick, CString sTo, CString sMsg, CString sFromNick )
{
	int i=-1;
	CString s;

	if ( IsUserOnline( sTo ) || CConfig::Instance()->GetChatSendOfflineMessages() )
	{
#ifdef HAVE_SSL
		s = m_pMessageSSL.EncryptMessage( this, sTo, sMsg );

		if ( s != "" )
		{
			sMsg = s;
		}
#endif

		i = CDCProto::SendPrivateMessage( sNick, sTo, sMsg, sFromNick );
	}

	return i;
}

/** */
int CClient::RequestNickList()
{
	// if m_nNickListHandler == 1 we are still waiting for the nicklist, TODO handling ???
	m_nNickListHandler = 1;
	m_nOpListHandler   = 1;

	return CDCProto::RequestNickList();
}

/** */
int CClient::Disconnect( bool force )
{
	UpdateReconnect( ersDISABLED );

	return CConnection::Disconnect(force);
}
