/***************************************************************************
                           cfilemanager.cpp  -  description
                             -------------------
    begin                : Sat Aug 03 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef WIN32
#include <unistd.h>
#include <stdlib.h>
#endif

#include <string.h>

#include <dclib/dcos.h>
//#include <dclib/cmd4.h>
#include <dclib/cmd5.h>
#include <dclib/cbytearray.h>
#include <dclib/cconfig.h>
#include <dclib/cstring.h>
#include <dclib/cstringlist.h>
#include <dclib/cdir.h>
#include <dclib/cfile.h>
#include <dclib/cdownloadmanager.h>

#include "cfilemanager.h"

class CFileType : public CObject {
public:
	/** */
	CFileType() {};
	/** */
	virtual ~CFileType() {};

	/** */
	eFileTypes m_eFileType;
};

/** filetype list */
typedef struct filetype {
	const char* sExt;
	eFileTypes eFileType;
} filetype;

struct filetype FileType[] = {
	// MP3 2
	{ "AIFF", eftMP3 },
	{ "AU",   eftMP3 },
	{ "MID",  eftMP3 },
	{ "MP2",  eftMP3 },
	{ "MP3",  eftMP3 },
	{ "OGG",  eftMP3 },
	{ "RA",   eftMP3 },
	{ "WAV",  eftMP3 },

	// ARCHIVE 3
	{ "ACE", eftARCHIVE },
	{ "BZ2", eftARCHIVE },
	{ "CAB", eftARCHIVE },
	{ "EX_", eftARCHIVE },
	{ "GZ",  eftARCHIVE },
	{ "LZH", eftARCHIVE },
	{ "RAR", eftARCHIVE },
	{ "RPM", eftARCHIVE },
	{ "TAR", eftARCHIVE },
	{ "TGZ", eftARCHIVE },
	{ "ZIP", eftARCHIVE },
	{ "ZOO", eftARCHIVE },
	{ "Z",   eftARCHIVE },

	// DOCUMENT 4
	{ "CFG",   eftDOCUMENT },
	{ "CPP",   eftDOCUMENT },
	{ "DIZ",   eftDOCUMENT },
	{ "DOC",   eftDOCUMENT },
	{ "H",     eftDOCUMENT },
	{ "HLP",   eftDOCUMENT },
	{ "HTM",   eftDOCUMENT },
	{ "HTML",  eftDOCUMENT },
	{ "INI",   eftDOCUMENT },
	{ "INF",   eftDOCUMENT },
	{ "LOG",   eftDOCUMENT },
	{ "PHP",   eftDOCUMENT },
	{ "PS",    eftDOCUMENT },
	{ "PDF",   eftDOCUMENT },
	{ "SHTML", eftDOCUMENT },
	{ "TXT",   eftDOCUMENT },
	{ "RFT",   eftDOCUMENT },
	{ "XML",   eftDOCUMENT },

	// APPL 5
	{ "BAT",  eftAPPLICATION },
	{ "CGI",  eftAPPLICATION },
	{ "COM",  eftAPPLICATION },
	{ "DLL",  eftAPPLICATION },
	{ "EXE",  eftAPPLICATION },
	{ "HQX",  eftAPPLICATION },
	{ "JS",   eftAPPLICATION },
	{ "SO",   eftAPPLICATION },
	{ "SYS",  eftAPPLICATION },
	{ "VXD",  eftAPPLICATION },

	// PICTURE 6
	{ "3DS",  eftPICTURE },
	{ "A11",  eftPICTURE },
	{ "ACB",  eftPICTURE },
	{ "ADC",  eftPICTURE },
	{ "ADI",  eftPICTURE },
	{ "AFI",  eftPICTURE },
	{ "AI",   eftPICTURE },
	{ "AIS",  eftPICTURE },
	{ "ANS",  eftPICTURE },
	{ "ART",  eftPICTURE },
	{ "B8",   eftPICTURE },
	{ "BMP",  eftPICTURE },
	{ "CBM",  eftPICTURE },
	{ "GIF",  eftPICTURE },
	{ "ICO",  eftPICTURE },
	{ "JPEG", eftPICTURE },
	{ "JPG",  eftPICTURE },
	{ "PIC",  eftPICTURE },
	{ "PNG",  eftPICTURE },
	{ "PCX",  eftPICTURE },
	{ "TGA",  eftPICTURE },
	{ "TIF",  eftPICTURE },

	// VIDEO 7
	{ "AVI",   eftVIDEO },
	{ "ASF",   eftVIDEO },
	{ "ASX",   eftVIDEO },
	{ "DAT",   eftVIDEO },
	{ "MOV",   eftVIDEO },
	{ "MOVIE", eftVIDEO },
	{ "MPG",   eftVIDEO },
	{ "MP4",   eftVIDEO },
	{ "MPEG",  eftVIDEO },
	{ "OGM",   eftVIDEO },
	{ "QT",    eftVIDEO },
	{ "RM",    eftVIDEO },
	{ "VIV",   eftVIDEO },
	{ "WMV",   eftVIDEO },
	{ "DIVX",  eftVIDEO },

	// END
	{ 0, eftALL }
};

/** */
CFileManager::CFileManager()
{
	CFile f;
	CDir d;
	CString s;
	ulonglong filesize;
	bool err = FALSE;

	m_pFileTypeList   = new CStringList();
	m_pFileNameList   = 0;
	m_nFileBaseIndex  = 0;

	m_pBaseArray       = new CByteArray();
	m_pFileBaseArray   = new CByteArray();
	m_pPathBaseArray   = new CByteArray();
	m_pSearchIndex     = new CByteArray();
	m_pSearchFileIndex = new CByteArray();
	m_pSearchBase      = new CByteArray();
	//m_pSearchArray     = 0;

	bCreateIndex     = FALSE;
	bCreateHashList  = FALSE;

	FileManagerInfo = new CFileManagerInfo();
	FileManagerInfo->m_bCreateShareList   = FALSE;
	FileManagerInfo->m_bCreateSearchIndex = FALSE;
	FileManagerInfo->m_nShareListPercent  = 100;

	InitFileTypeList();

	// load the filelist
	s = CConfig::Instance()->GetConfigPath()+"database.bin";
	filesize = d.getFileSize(s,FALSE);

	if ( (filesize%sizeof(struct filebaseobject)) != 0 )
	{
		err = TRUE;
	}

	if ( err == FALSE )
	{
		if ( f.Open( s, IO_RAW | IO_READONLY ) == TRUE )
		{
			m_pBaseArray->SetSize(filesize);

			if ( f.Read( (char*)m_pBaseArray->Data(), m_pBaseArray->Size() ) != filesize )
			{
				err = TRUE;
			}

			f.Close();
		}
		else
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		s = CConfig::Instance()->GetConfigPath()+"filebase.bin";
		filesize = d.getFileSize(s,FALSE);

		if ( f.Open( s, IO_RAW | IO_READONLY ) == TRUE )
		{
			m_pFileBaseArray->SetSize(filesize);

			if ( f.Read( (char*)m_pFileBaseArray->Data(), m_pFileBaseArray->Size() ) != filesize )
			{
				err = TRUE;
			}

			f.Close();
		}
		else
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		s = CConfig::Instance()->GetConfigPath()+"pathbase.bin";
		filesize = d.getFileSize(s,FALSE);

		if ( f.Open( s,	IO_RAW | IO_READONLY ) == TRUE )
		{
			m_pPathBaseArray->SetSize(filesize);

			if ( f.Read( (char*)m_pPathBaseArray->Data(), m_pPathBaseArray->Size() ) != filesize )
			{
				err = TRUE;
			}

			f.Close();
		}
		else
		{
			err = TRUE;
		}
	}

	if ( err )
	{
		m_pBaseArray->SetSize(0);
		m_pFileBaseArray->SetSize(0);
		m_pPathBaseArray->SetSize(0);
	}

	if ( err == FALSE )
	{
		s = CConfig::Instance()->GetConfigPath()+"searchbase.bin";
		filesize = d.getFileSize(s,FALSE);

		if ( f.Open( s,	IO_RAW | IO_READONLY ) == TRUE )
		{
			m_pSearchBase->SetSize(filesize);

			if ( f.Read( (char*)m_pSearchBase->Data(), m_pSearchBase->Size() ) != filesize )
			{
				err = TRUE;
			}

			f.Close();
		}
		else
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		s = CConfig::Instance()->GetConfigPath()+"searchindex.bin";
		filesize = d.getFileSize(s,FALSE);

		if ( f.Open( s,	IO_RAW | IO_READONLY ) == TRUE )
		{
			m_pSearchIndex->SetSize(filesize);

			if ( f.Read( (char*)m_pSearchIndex->Data(), m_pSearchIndex->Size() ) != filesize )
			{
				err = TRUE;
			}

			f.Close();
		}
		else
		{
			err = TRUE;
		}
	}

	if ( err == FALSE )
	{
		s = CConfig::Instance()->GetConfigPath()+"searchfileindex.bin";
		filesize = d.getFileSize(s,FALSE);

		if ( f.Open( s,	IO_RAW | IO_READONLY ) == TRUE )
		{
			m_pSearchFileIndex->SetSize(filesize);

			if ( f.Read( (char*)m_pSearchFileIndex->Data(), m_pSearchFileIndex->Size() ) != filesize )
			{
				err = TRUE;
			}

			f.Close();
		}
		else
		{
			err = TRUE;
		}
	}

	if ( err )
	{
		m_pSearchBase->SetSize(0);
		m_pSearchIndex->SetSize(0);
		m_pSearchFileIndex->SetSize(0);
		m_nSearchFileIndex = 0;

		// TODO: recreate index
	}
	else
	{
		m_nSearchFileIndex = (struct long_s*)m_pSearchFileIndex->Data();
	}

	SetInstance(this);
}

/** */
CFileManager::~CFileManager()
{
	Stop();

	Lock();

	if ( m_pFileNameList )
	{
		delete m_pFileNameList;
		m_pFileNameList = 0;
	}

	if ( FileManagerInfo )
	{
		delete FileManagerInfo;
		FileManagerInfo = 0;
	}

	delete m_pFileTypeList;
	delete m_pBaseArray;
	delete m_pFileBaseArray;
	delete m_pPathBaseArray;
	delete m_pSearchIndex;
	delete m_pSearchBase;

	UnLock();
}

/** */
void CFileManager::InitFileTypeList()
{
	int i;
	CFileType * filetype;

	for(i=0;FileType[i].sExt != 0;i++)
	{
		filetype = new CFileType();
		filetype->m_eFileType = FileType[i].eFileType;

		m_pFileTypeList->Add( FileType[i].sExt, filetype );
	}
}

/** return the filetype */
eFileTypes CFileManager::GetFileType( CString file )
{
	CString ext;
	eFileTypes res=eftUNKNOWN;
	CFileType * filetype;

	// set the filetype
	ext = CDir::Extension(file);

	if ( ext != "" )
	{
		if ( m_pFileTypeList->Get( ext.ToUpper(), (CObject*&)filetype ) == 0 )
		{
			res = filetype->m_eFileType;
		}
#if 0
		if ( res == eftUNKNOWN )
			printf("unknown extension: %s\n",ext.Data());
#endif
	}

	return res;
}

/** */
CStringList * CFileManager::Search( CString s )
{
	CString *ps;
	CStringList *sl=0;
	struct fileindexobject *pfio;
	struct searchindexobject * psio;

	if ( !FileManagerInfo )
	{
		return 0;
	}

	if ( (FileManagerInfo->m_bCreateShareList == TRUE) ||
	     (FileManagerInfo->m_bCreateSearchIndex == TRUE) )
	{
		return 0;
	}

	Lock();

	if ( (psio = search_index(s) ) != 0 )
	{
		sl = new CStringList();

		pfio = (struct fileindexobject*) (m_pSearchIndex->Data()+psio->m_nFileIndex);

		while(pfio)
		{
			ps = new CString( CString().setNum(pfio->m_nFileBaseIndex));
			sl->Add( *ps, ps );

			if ( pfio->m_nNext == 0 )
			{
				pfio = 0;
			}
			else
			{
				pfio = (struct fileindexobject*) (m_pSearchIndex->Data()+pfio->m_nNext);
			}
		}
	}

	UnLock();

	return sl;
}

/** */
bool CFileManager::GetFileBaseObject( CString id, struct filebaseobject * fbo, CString & filename )
{
	bool res = FALSE;
	ulonglong index;

	if ( !FileManagerInfo )
	{
		return res;
	}

	if ( (FileManagerInfo->m_bCreateShareList == TRUE) ||
	     (FileManagerInfo->m_bCreateSearchIndex == TRUE) )
	{
		return res;
	}

	Lock();

	index = id.asULL();

	if ( (index*sizeof(struct filebaseobject)) < m_pBaseArray->Size() )
	{
		memcpy( fbo, m_pBaseArray->Data()+(index*sizeof(struct filebaseobject)), sizeof(struct filebaseobject) );

		if ( m_pPathBaseArray->Size() > fbo->m_nPathIndex )
		{
			filename = (char*)m_pPathBaseArray->Data()+fbo->m_nPathIndex;

			if ( filename != "" )
			{
				filename += DIRSEPARATOR;
			}
		}

		if ( m_pFileBaseArray->Size() > fbo->m_nFileIndex )
		{
			filename += (char*)m_pFileBaseArray->Data()+fbo->m_nFileIndex;
		}

		res = TRUE;
	}

	UnLock();

	return res;
}

/** */
void CFileManager::SaveIndex()
{
	CFile f;

	if ( !FileManagerInfo )
	{
		return;
	}

	if ( f.Open( CConfig::Instance()->GetConfigPath()+"searchbase.bin", 
			IO_RAW | IO_WRITEONLY | IO_TRUNCATE | IO_CREAT,
			MO_IRUSR | MO_IWUSR ) == TRUE )
	{
		f.Write( (const char*)m_pSearchBase->Data(), m_pSearchBase->Size() );
		f.Close();
	}

	if ( f.Open( CConfig::Instance()->GetConfigPath()+"searchindex.bin", 
			IO_RAW | IO_WRITEONLY | IO_TRUNCATE | IO_CREAT,
			MO_IRUSR | MO_IWUSR ) == TRUE )
	{
		f.Write( (const char*)m_pSearchIndex->Data(), m_pSearchIndex->Size() );
		f.Close();
	}
	
	if ( f.Open( CConfig::Instance()->GetConfigPath()+"searchfileindex.bin", 
			IO_RAW | IO_WRITEONLY | IO_TRUNCATE | IO_CREAT,
			MO_IRUSR | MO_IWUSR ) == TRUE )
	{
		f.Write( (const char*)m_pSearchFileIndex->Data(), m_pSearchFileIndex->Size() );
		f.Close();
	}

	if ( f.Open( CConfig::Instance()->GetConfigPath()+"database.bin", 
			IO_RAW | IO_WRITEONLY | IO_TRUNCATE | IO_CREAT,
			MO_IRUSR | MO_IWUSR ) == TRUE )
	{
		f.Write( (const char*)m_pBaseArray->Data(), m_pBaseArray->Size() );
		f.Close();
	}

	if ( f.Open( CConfig::Instance()->GetConfigPath()+"filebase.bin", 
			IO_RAW | IO_WRITEONLY | IO_TRUNCATE | IO_CREAT,
			MO_IRUSR | MO_IWUSR ) == TRUE )
	{
		f.Write( (const char*)m_pFileBaseArray->Data(), m_pFileBaseArray->Size() );
		f.Close();
	}

	if ( f.Open( CConfig::Instance()->GetConfigPath()+"pathbase.bin", 
			IO_RAW | IO_WRITEONLY | IO_TRUNCATE | IO_CREAT,
			MO_IRUSR | MO_IWUSR ) == TRUE )
	{
		f.Write( (const char*)m_pPathBaseArray->Data(), m_pPathBaseArray->Size() );
		f.Close();
	}
}

/** */
void CFileManager::StartIndex()
{
}

/** */
void CFileManager::StopIndex()
{
	SaveIndex();
}

/** */
void CFileManager::AddIndex( CFileInfo *fileinfo, CString path, CString /*rootpath*/ )
{
	struct filebaseobject fbo;

	fbo.m_eFileType  = GetFileType(fileinfo->name);
	fbo.m_nFileIndex = m_pFileBaseArray->Size();
	fbo.m_nPathIndex = m_pPathBaseArray->Size();
	fbo.m_nSize      = fileinfo->size;
	fbo.m_tModTime   = fileinfo->st_m_time;

	m_pBaseArray->Append( (const char*)&fbo, sizeof(struct filebaseobject) );
	m_pFileBaseArray->Append( fileinfo->name.Data(), fileinfo->name.Length()+1 );
	m_pPathBaseArray->Append( path.Data(), path.Length()+1 );
}

/** */
bool CFileManager::CreateShareList()
{
	bool err = FALSE;

	Lock();

	if ( (FileManagerInfo->m_bCreateSearchIndex == FALSE) &&
	     (FileManagerInfo->m_bCreateShareList == FALSE) )
	{
		SharedFolders.Clear();

		pShareFolder     = 0;
		iShareSize       = 0;
		sShareBuffer     = "";

		m_pBaseArray->SetSize(0);
		m_pFileBaseArray->SetSize(0);
		m_pPathBaseArray->SetSize(0);

		if ( CConfig::Instance()->GetSharedFolders(&SharedFolders) != 0 )
		{
			if ( m_pFileNameList )
				delete m_pFileNameList;
			m_pFileNameList  = new CStringList(25);

			FileManagerInfo->m_nShareListPercent = 0;
			FileManagerInfo->m_bCreateShareList  = TRUE;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);

			err = TRUE;
		}
		else
		{
			CConfig::Instance()->SetShareBuffer( "", 0 );

			SaveIndex();
		}
	}

	if ( err == TRUE )
	{	
		StartIndex();

		Start();
	}

	UnLock();

	return err;
}

/** */
void CFileManager::CreateSearchIndex()
{
	if ( (FileManagerInfo->m_bCreateSearchIndex == FALSE) &&
	     (FileManagerInfo->m_bCreateShareList == FALSE) )
	{
		m_nFileBaseIndex = 0;
		m_nSearchIndex   = 0;

		m_pSearchIndex->SetSize(0);
		m_pSearchFileIndex->SetSize(0);
		m_pSearchBase->SetSize(0);

		//m_pSearchArray = new CByteArray*[0x100];
		memset(m_pSearchArray,0,sizeof(m_pSearchArray));

		FileManagerInfo->m_nSearchIndexPercent = 0;
		FileManagerInfo->m_bCreateSearchIndex  = TRUE;

		Start();
	}
}

/** */
bool CFileManager::CreateHashList()
{
	return FALSE;
}

struct searchindexobject * CFileManager::search_index( CString & s )
{
	unsigned long i,is,ie;
	unsigned int in = s.Data()[0]&0xFF;
	struct searchindexobject * psio = 0;

	if ( (!m_nSearchFileIndex) || (m_nSearchFileIndex[in].n == 0) )
	{
		return 0;
	}

	// set start
	is = m_nSearchFileIndex[in].n;

	// find end
	for(ie=0,i=in+1;i<=0xFF;i++)
	{
		if ( m_nSearchFileIndex[i].n != 0 )
		{
			ie = m_nSearchFileIndex[i].n;
			break;
		}
	}

	if ( ie == 0 )
	{
		ie = m_pSearchFileIndex->Size();
	}

	i = 0;

	while( (is+(i*sizeof(struct searchindexobject))) < ie )
	{
		psio = (struct searchindexobject *) (m_pSearchFileIndex->Data()+(is+(i*sizeof(struct searchindexobject))));

		if ( s == (char*)(m_pSearchBase->Data()+psio->m_nIndex) )
		{
			break;
		}

		i++;
		psio = 0;
	}

	return psio;
}

struct searchindexobject * CFileManager::find_index( CString & s )
{
	long i=0;
	unsigned int in = s.Data()[0]&0xFF;
	struct searchindexobject * psio = 0;

	if ( m_pSearchArray[in] == 0 )
	{
		return 0;
	}

	while( (i*sizeof(struct searchindexobject)) < m_pSearchArray[in]->Size() )
	{
		psio = (struct searchindexobject *) (m_pSearchArray[in]->Data()+(i*sizeof(struct searchindexobject)));

		if ( s == (char*)(m_pSearchBase->Data()+psio->m_nIndex) )
		{
			break;
		}

		i++;
		psio = 0;
	}

	return psio;
}

void CFileManager::add_index( CString & s )
{
	unsigned int in;
	struct searchindexobject sio, *psio;
	struct fileindexobject fio, *pfio;

	in = s.Data()[0]&0xFF;
			
	if ( m_pSearchArray[in] == 0 )
	{
		m_pSearchArray[in] = new CByteArray();
	}
			
	if ( (psio = find_index(s) ) != 0 )
	{
		pfio = (struct fileindexobject*) (m_pSearchIndex->Data()+psio->m_nFileIndex);

		while(pfio->m_nNext!=0)
		{
			pfio = (struct fileindexobject*) (m_pSearchIndex->Data()+pfio->m_nNext);
		}

		pfio->m_nNext = m_pSearchIndex->Size();
	}
	else
	{
		sio.m_nIndex     = m_pSearchBase->Size();
		sio.m_nFileIndex = m_pSearchIndex->Size();

		m_pSearchBase->Append(s.Data(),s.Length()+1);
		m_pSearchArray[in]->Append( (const char*)&sio, sizeof(struct searchindexobject) );
	}

	fio.m_nFileBaseIndex = m_nFileBaseIndex;
	fio.m_nNext          = 0;

	m_pSearchIndex->Append( (const char*)&fio, sizeof(struct fileindexobject) );
}

/** */
void CFileManager::ThreadCreateSearchIndex()
{
	int i,k;
	CString s,s1,id;
//	CStringList *sl;
//	CString * ps;
	struct filebaseobject * fbo;
	double percent;

	for(k=0;k<100;k++)
	{
		if ( (m_nFileBaseIndex*sizeof(struct filebaseobject)) < m_pBaseArray->Size() )
		{
			fbo = (struct filebaseobject *) (m_pBaseArray->Data()+(m_nFileBaseIndex*sizeof(struct filebaseobject)));

			m_nSearchIndex++;

			//id = CString().setNum(m_nFileBaseIndex);

			// get file
			s = (char*) (m_pFileBaseArray->Data()+fbo->m_nFileIndex);

			s = s.ToUpper();

			add_index(s);

			// convert special chars
			s = s.Replace(' ',"/");
			s = s.Replace('.',"/");
			s = s.Replace('-',"/");
			s = s.Replace('_',"/");
			s = s.Replace('(',"/");
			s = s.Replace(')',"/");
			s = s.Replace('!',"/");

			s = CDir().ConvertSeparators(s);

			while( (i=s.Find(DIRSEPARATOR)) != -1 )
			{
				s1 = s.Left(i);
				s  = s.Mid(i+1,s.Length()-i-1);

				i++;

				// max length is 3
				if ( s1.Length() < 3 )
					continue;

				add_index(s1);
			}

			m_nFileBaseIndex++;
		}
		else
		{
			// init array
			m_pSearchFileIndex->SetSize(sizeof(struct long_s)*0x100);

			for(i=0;i<=0xFF;i++)
			{
				m_nSearchFileIndex = (struct long_s*)m_pSearchFileIndex->Data();

				m_nSearchFileIndex[i].n = 0;

				if ( m_pSearchArray[i] != 0 )
				{
					m_nSearchFileIndex[i].n = m_pSearchFileIndex->Size();
					m_pSearchFileIndex->Append(m_pSearchArray[i]->Data(),m_pSearchArray[i]->Size());
					delete m_pSearchArray[i];
				}
			}

			//delete m_pSearchArray;
			//m_pSearchArray = 0;

			printf("ready create index\n");

			SaveIndex();

			FileManagerInfo->m_nSearchIndexPercent = 100;
			FileManagerInfo->m_bCreateSearchIndex  = FALSE;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);

			Stop();

			return;
		}
	}

	if ( m_pBaseArray->Size() > 0 )
	{
		percent = (m_nFileBaseIndex*100.0)/(m_pBaseArray->Size()/sizeof(struct filebaseobject));

		if ( FileManagerInfo->m_nSearchIndexPercent != percent )
		{
			FileManagerInfo->m_nSearchIndexPercent = percent;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);
		}
	}
}

/** */
void CFileManager::ThreadCreateShareList()
{
	CString s,s1;
	CDir dir;

	if ( (pShareFolder=SharedFolders.Next(pShareFolder)) != 0 )
	{
		// directory
		s = pShareFolder->m_sPath;

		// is valid ?
		if ( dir.cd(s.Data()) == TRUE )
		{
			s  = dir.Path();
			s1 = dir.DirName();

			if ( s1 != "" )
			{
				s = s.Left( s.Length()-s1.Length() );
			}

			ThreadCreateShareList( 0, s, s1, "", 100/SharedFolders.Count() );
		}
		else
		{
			printf("Can't change dir: '%s'\n",s.Data());
		}
	}
	else
	{
		FileManagerInfo->m_nShareListPercent = 100;
		printf("ready create sharelist\n");

		CConfig::Instance()->SetShareBuffer( sShareBuffer, iShareSize );

		if ( m_pFileNameList )
		{
			delete m_pFileNameList;
			m_pFileNameList = 0;
		}

		FileManagerInfo->m_bCreateShareList = FALSE;

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);

		StopIndex();

		CreateSearchIndex();

		if ( CDownloadManager::Instance() )
			CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);
	}
}

/** */
void CFileManager::ThreadCreateShareList( int depth, CString base, CString curr, CString relpath, double percent )
{
	CDir dir;
	CList<CFileInfo> list;
	CString d,s,*ts=0;
	CString path;
	CString rpath;
	CFileInfo *fileinfo;
	int x,z;

	if ( relpath == "" )
		rpath = curr;
	else
		rpath = relpath+DIRSEPARATOR+curr;

	if (base=="")
		path = curr;
	else
		path = base+DIRSEPARATOR+curr;

	dir.SetPath(path);

	d = "";
	z = depth;

//	printf("%s\n",curr.Data());

	// create depth string
	if ( curr != "" )
	{
		for(x=depth;x>0;x--) d+="\x9";
		z++;
		// use alias
		if ( depth == 0 )
			sShareBuffer += pShareFolder->m_sAlias + "\xd\xa";
		else
			sShareBuffer += d + curr + "\xd\xa";
		d += "\x9";
	}

	// get all files
	if ( dir.ReadEntrys( CDir::Files, &list ) )
	{
		// add files
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( fileinfo->name != "" )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					if ( m_pFileNameList->Get(s,(CObject*&)ts) != 0 )
					{
						m_pFileNameList->Add(s,new CString());
						iShareSize   += fileinfo->size;
						sShareBuffer += d + fileinfo->name + "|" + CString().setNum(fileinfo->size) + "\xd\xa";
					}
				}

				// add file to the file manager ...
				AddIndex( fileinfo, rpath, path );
			}
		}
	}

	// get all dirs
	if ( dir.ReadEntrys( CDir::Dirs, &list ) )
	{
		if ( list.Count() > 0 )
			percent /= list.Count();
		else
			FileManagerInfo->m_nShareListPercent += percent;

		// add dir
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( (fileinfo->name != ".") && (fileinfo->name != "..") )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					if ( m_pFileNameList->Get(s,(CObject*&)ts) != 0 )
					{
						m_pFileNameList->Add(s,new CString());
						ThreadCreateShareList(z,path,fileinfo->name,rpath,percent);
					}
					else
						FileManagerInfo->m_nShareListPercent += percent;

				}
			}
			else
				FileManagerInfo->m_nShareListPercent += percent;

			if ( CDownloadManager::Instance() )
				CDownloadManager::Instance()->SendFileManagerInfo(FileManagerInfo);
		}
	}
}

/** */
ulonglong CFileManager::GetShareSize()
{
	CString s,s1;
	CDir dir;
	ulonglong iSSize;

	iSSize = 0;
	pShareFolder=0;

	if ( CConfig::Instance()->GetSharedFolders(&SharedFolders) == 0 )
	{
		printf("No share folderfound !");
		return 0;
	}

	while ( (pShareFolder=SharedFolders.Next(pShareFolder)) != 0 )
	{
		// directory
		s = pShareFolder->m_sPath;

		// is valid ?
		if ( dir.cd(s.Data()) == TRUE )
		{
			s  = dir.Path();
			s1 = dir.DirName();

			if ( s1 != "" )
			{
				s = s.Left( s.Length()-s1.Length() );
			}

			iSSize += GetShareSize( 0, s, s1, "" );
		}
		else
		{
			printf("Can't change to dir: '%s'\n",s.Data());
		}
	}

	return iSSize;
}


/** */
ulonglong CFileManager::GetShareSize( int depth, CString base, CString curr, CString relpath )
{
	ulonglong lShareSize;
	CDir dir;
	CList<CFileInfo> list;
	CString d,s;
	CString path;
	CString rpath;
	CFileInfo *fileinfo;
	int x,z;

	lShareSize = 0;

	if ( relpath == "" )
		rpath = curr;
	else
		rpath = relpath+DIRSEPARATOR+curr;

	if (base=="")
		path = curr;
	else
		path = base+DIRSEPARATOR+curr;

	dir.SetPath(path);

	d = "";
	z = depth;

	// create depth string
	if ( curr != "" )
	{
		for(x=depth;x>0;x--) d+="\x9";
		z++;
		d += "\x9";
	}

	// get all files
	if ( dir.ReadEntrys( CDir::Files, &list ) )
	{
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( fileinfo->name != "" )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					lShareSize += fileinfo->size;
				}
			}
		}
	}

	// get all dirs
	if ( dir.ReadEntrys( CDir::Dirs, &list ) )
	{
		fileinfo = 0;
		while( (fileinfo=list.Next(fileinfo)) != 0 )
		{
			if ( (fileinfo->name != ".") && (fileinfo->name != "..") )
			{
				if ( fileinfo->m_bSymlink )
					s = dir.ReadLink(path+DIRSEPARATOR+fileinfo->name);
				else
					s = dir.SimplePath(path+DIRSEPARATOR+fileinfo->name);

				if ( s != "" )
				{
					lShareSize += GetShareSize(z,path,fileinfo->name,rpath);
				}
			}
		}
	}
	return lShareSize;
}

/** */
void CFileManager::ThreadCreateHashList()
{
/*
	CFileBaseObject * FileBaseObject = 0;
	CString s;
	FILE * file;
	CMD5 * md5;
	MD4 * md4;
	int hashes;
	unsigned char * buffer, *phashes;
	int len;
	int i,j;

	while ( pFileBaseList->Next( (CObject*&)FileBaseObject) )
	{
		if ( (FileBaseObject->sMD5Hash == "") &&
		     (FileBaseObject->lSize > 0) )
		{
			// create the hash
			s = FileBaseObject->sRootPath + DIRSEPARATOR + FileBaseObject->sFile;

			printf("CREATE HASH: %s\n",s.Data());

			if ( (file=fopen(s.Data(),"rb")) != 0 )
			{
				hashes  = (FileBaseObject->lSize / EDONKEY_HASH_BLOCK_SIZE) + 1;
				phashes = (unsigned char *) malloc(hashes*16);
				buffer  = (unsigned char *) malloc(EDONKEY_HASH_BLOCK_SIZE/10);

//				printf("create %d md4 chunk hashes %llu\n",hashes,FileBaseObject->lSize);
				md5 = new CMD5();

				i = j = 0;
				md4 = 0;

				while ( !Stopped() )
				{
					len = fread(buffer, 1, EDONKEY_HASH_BLOCK_SIZE/10, file);

//					printf("%d %d : %d\n",i,j,len);

					if ( md4 == 0 )
					{
						md4 = new MD4();
					}

					if ( len > 0 )
					{
						md4->update(buffer, len);
						md5->update(buffer, len);
					}

					j++;

					if ( (j == 10) || (len == 0) )
					{
						md4->finalize();
						md4->raw_digest( phashes + (i*16) );

						delete md4;
						md4 = 0;
						i++;
						j = 0;
					}

					if ( (len == 0) || (i == hashes) )
					{
						break;
					}

					NanoSleep(25);
				}

//				printf("ready %d %d\n",i,hashes);

				if ( !Stopped() )
				{
					if ( hashes > 1 )
					{
						md4 = new MD4();
						md4->update(phashes, hashes*16);
						md4->finalize();
						FileBaseObject->sMD4Hash = md4->hex_digest();

						for(i=0;i<hashes;i++)
						{
							s = md4->hex_digest( phashes + (i*16) );
							FileBaseObject->pMD4ChunkHashList.Add( new CString(s) );
						}

						delete md4;
					}
					else
					{
						s = md4->hex_digest( phashes );
						FileBaseObject->sMD4Hash = s;
						FileBaseObject->pMD4ChunkHashList.Add( new CString(s) );
					}

					md5->finalize();

					FileBaseObject->sMD5Hash = md5->hex_digest();
				}

				delete md5;

				free(buffer);
				free(phashes);

				fclose(file);
				break;
			}
			else
			{
				printf("ERROR: %s\n",FileBaseObject->sFile.Data());
			}
		}
	}
*/
/*
	if ( (FileBaseObject == 0) && (!Stopped()) )
	{
		bCreateIndex = FALSE;
		Stop();
	}
*/
}

/** */
void CFileManager::Thread( CObject * )
{
	int wait = 50;

	Lock();

	if ( FileManagerInfo->m_bCreateShareList == TRUE )
	{
		ThreadCreateShareList();
	}
	else if ( FileManagerInfo->m_bCreateSearchIndex == TRUE )
	{
		ThreadCreateSearchIndex();

		wait = 0;
	}

	UnLock();

	NanoSleep(wait);
}
