/***************************************************************************
                           cfilemanager.h  -  description
                             -------------------
    begin                : Sat Aug 03 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CFILEMANAGER_H
#define CFILEMANAGER_H

/**
  *@author Mathias Kster
  */

#include <dclib/dcos.h>
#include <dclib/cobject.h>
#include <dclib/cthread.h>
#include <dclib/cmessagehandler.h>
#include <dclib/csingleton.h>
#include <dclib/cconfig.h>

class CStringList;
class CThreadStringList;
class CFileInfo;
class CFileManagerInfo;

#pragma pack(1)

typedef struct filebaseobject {
	/** */
	unsigned int m_eFileType;
	/** */
	ulonglong m_nSize;
	/** */
	unsigned long m_nPathIndex;
	/** */
	unsigned long m_nFileIndex;
	/** */
	time_t m_tModTime;
} filebaseobject;

typedef struct searchindexobject {
	/** */
	unsigned long m_nIndex;
	/** */
	unsigned long m_nFileIndex;
} searchindexobject;

typedef struct fileindexobject {
	/** */
	ulonglong m_nFileBaseIndex;
	/** */
	unsigned long m_nNext;
} fileindexobject;

typedef struct long_s {
	/** */
	unsigned long n;
} long_s;

#pragma pack()

class CFileManager;
class CByteArray;

#ifdef WIN32
template class DLL_EXPORT CSingleton<CFileManager>;
#endif

class DLL_EXPORT CFileManager : public CSingleton<CFileManager>, public CThread {

public:
	/** */
	CFileManager();
	/** */
	virtual ~CFileManager();

	/** */
	void SaveIndex();
	/** */
	void StartIndex();
	/** */
	void StopIndex();
	/** */
	void AddIndex( CFileInfo * fileinfo, CString path, CString rootpath );
	/** */
	CStringList * Search( CString s );
	/** */
	bool GetFileBaseObject( CString id, struct filebaseobject * fbo, CString & filename );

	/** */
	bool CreateShareList();
	/** */
	bool CreateHashList();
	/** */
	void CreateSearchIndex();

	/** */
	ulonglong GetShareSize();

	/** thread callbackfunction */
	virtual void Thread( CObject * Object );

private:
	/** */
	void ThreadCreateSearchIndex();
	/** */
	void ThreadCreateHashList();
	/** */
	void ThreadCreateShareList();
	/** */
	void ThreadCreateShareList( int depth, CString base, CString curr, CString relpath, double percent );

	/** */
	ulonglong GetShareSize( int depth, CString base, CString curr, CString relpath );
	/** */
	void InitFileTypeList();

	struct searchindexobject * find_index( CString & s );
	struct searchindexobject * search_index( CString & s );
	void add_index( CString & s );

	/** */
	CByteArray * m_pFileBaseArray;
	/** */
	CByteArray * m_pPathBaseArray;
	/** */
	CByteArray * m_pBaseArray;
	/** */
	CByteArray * m_pSearchIndex;
	/** */
	CByteArray * m_pSearchFileIndex;
	/** */
	CByteArray * m_pSearchBase;
	/** */
	CByteArray * m_pSearchArray[0x100];
	/** */
	struct long_s * m_nSearchFileIndex;

	/** */
	CStringList * m_pFileNameList;
	/** */
	CStringList * m_pFileTypeList;

	/** return the filetype */
	eFileTypes GetFileType( CString file );

	/** */
	CList<DCConfigShareFolder> SharedFolders;
	/** */
	DCConfigShareFolder * pShareFolder;
	/** */
	ulonglong m_nFileBaseIndex;
	/** */
	CString sShareBuffer;
	/** */
	ulonglong iShareSize;

	/** */
	bool bCreateHashList;
	/** */
	bool bCreateIndex;

	/** */
	long m_nSearchIndex;

	/** */
	CFileManagerInfo * FileManagerInfo;
};

#endif
