/***************************************************************************
                           cmanager.cpp  -  description
                             -------------------
    begin                : Thu Jan 23 2003
    copyright            : (C) 2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#ifndef WIN32
#include <sys/time.h>
#else
#include <time.h>
#include <sys/timeb.h> 
extern "C" { 
int gettimeofday(struct timeval *tp, void * );
}
#endif
#include <dclib/dcos.h>
#include <dclib/cdir.h>
#include <dclib/cplugin.h>
#include <dclib/cconfig.h>
#include <dclib/cdownloadmanager.h>
#include <dclib/cfilemanager.h>
#include <dclib/cquerymanager.h>
#include <dclib/cservermanager.h>

#include "cmanager.h"

/** */
CManager::CManager()
{
	Start();

	SetInstance(this);
}

/** */
CManager::~CManager()
{
	Stop();
}

/** */
void CManager::Add( _CCallback * callback )
{
	if ( callback )
	{
		m_ThreadList.Lock();
		m_ThreadList.Add(callback);
		m_ThreadList.UnLock();
	}
}

/** */
void CManager::Remove( _CCallback * callback )
{
	if ( callback )
	{
		m_ThreadList.Lock();
		m_ThreadList.Remove(callback);
		m_ThreadList.UnLock();
	}
}

/** */
void CManager::Thread( CObject * )
{
	struct timeval tv_old, tv_new;
	_CCallback * callback = 0;
	int i;

	gettimeofday(&tv_old, NULL);

	while( TRUE )
	{
		m_ThreadList.Lock();

		callback = m_ThreadList.Next(callback);

		m_ThreadList.UnLock();

		if ( !callback )
		{
			break;
		}

		callback->notify(0,0);
	}

	gettimeofday(&tv_new, NULL);

	if ( (tv_new.tv_sec - tv_old.tv_sec) > 0 )
	{
		i = 0;
	}
	else
	{
		i = (tv_new.tv_usec - tv_old.tv_usec)/1000;

		if ( i > 50 )
		{
			i = 1;
		}
		else
		{
			i = 50-i;
		}
	}

	//malloc_stats();

	//printf("%d %d %d\n",i,tv_new.tv_sec - tv_old.tv_sec,(tv_new.tv_usec - tv_old.tv_usec)/1000);

	NanoSleep(i);
}

extern "C" {
// QpThread archive

#ifdef WIN32
/*
 * gettimeofday()
 */
int gettimeofday(struct timeval *tp, void * /*pv*/ )
{
#if 0
/*
 * ORIGINAL VERSION, that does not work
 *
  SYSTEMTIME st;
  struct tm  t;

  GetSystemTime(&st);

  memset((void *) &t, 0, sizeof(struct tm));
  // typecast is the best trick to handle negative valuest
  t.tm_sec  = (signed short ) st.wSecond;
  t.tm_min  = (signed short ) st.wMinute;
  t.tm_hour = (signed short ) st.wHour;
  t.tm_mday = (signed short ) st.wDay;
  t.tm_mon  = (signed short ) (st.wMonth - 1);
  t.tm_year = (signed short ) (st.wYear - 1900);
  t.tm_isdst = -1;

  tp->tv_sec  = mktime(&t);                   
  tp->tv_usec = st.wMilliseconds;
 *
 * NEW VERSION
 */
#endif

#if 0
	SYSTEMTIME st;
	GetSystemTime(&st);
	tp->tv_sec = time(NULL);
	tp->tv_usec = st.wMilliseconds;
#endif
	
        struct _timeb currSysTime;
        
        /* get current system time */
        _ftime(&currSysTime);
	
	tp->tv_sec = currSysTime.time;
	tp->tv_usec = currSysTime.millitm * 1000;

	return 0;
}

#endif
} /* extern "C" */
