/***************************************************************************
                          ctransfer.cpp  -  description
                             -------------------
    begin                : Fri Feb 22 2002
    copyright            : (C) 2002-2003 by Mathias Kster
    email                : mathen@ketelhot.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#ifndef WIN32
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#else
#include <io.h>
#endif

#include <dclib/dcos.h>
#include <dclib/dcobject.h>
#include <dclib/cmessagehandler.h>
#include <dclib/cencrypt.h>
#include <dclib/che3.h>
#include <dclib/cbz.h>
#include <dclib/cdir.h>
#include <dclib/cbytearray.h>
#include <dclib/cconfig.h>

#include "ctransfer.h"

#define SEND_FILE_BUFFER_SIZE	0xffff

CTransfer::CTransfer( bool listener )
{
	pCallback = 0;

	fd = -1;

	pByteArray = new CByteArray();

	m_bListener   = listener;
	sNick         = "";
	sSrcFilename  = "";
	sDstFilename  = "";
	sBuffer       = "";
	eMedium       = eltNONE;
	eSrcDirection = edNONE;
	eDstDirection = edNONE;
	eMode         = estNONE;

	lStartPosition   = 0;
	lCurrentPosition = 0;
	lEndPosition     = 0;
	m_nSrcLevel      = 0;
	m_nDstLevel      = 0;
	starttime        = 0;
	currtime         = 0;
	ratetime         = 0;
	ratetransfered   = 0;
	lLength          = 0;
	lTransfered      = 0;
	lRate            = 0;
	lChunkSize       = 0;
	m_nCurrRate      = 0;

	sDstNick = "";

	sHubName = "";
	sHubHost = "";

	sTransferID = "";

	m_bIdle = TRUE;
	m_eTransferState = etsNONE;

	m_eTransferType = ettNONE;
}

CTransfer::~CTransfer()
{
	SetCallBackFunction(0);

	TransferThread.Lock();

	if (pByteArray)
	{
		delete pByteArray;
		pByteArray = 0;
	}

	if ( fd != -1 )
	{
#ifdef WIN32
		_close(fd);
#else
		close(fd);
#endif
	}

	TransferThread.UnLock();
}

/** */
int CTransfer::CallBack_SendObject( CObject * Object )
{
	int err;

	if ( pCallback != 0 )
	{
		err = pCallback->notify( this, Object );
	}
	else
	{
		err = DC_CallBack( Object );
	}

	if ( err == -1 )
	{
		printf("CallBack failed (state)...\n");
		delete Object;
	}

	return err;
}

/** */
int CTransfer::CallBack_SendError( CString msg )
{
	CMessageError * Object = new CMessageError();
	Object->m_eType = DC_MESSAGE_ERROR;
	Object->sError  = msg;

	return CallBack_SendObject(Object);
}

/** */
void CTransfer::SetBuffer( CByteArray *ba )
{
	TransferThread.Lock();
	pByteArray->Append(ba->Data(),ba->Size());
	TransferThread.UnLock();
}

/** */
long CTransfer::GetBuffer( CByteArray *ba )
{
	TransferThread.Lock();
	long i;
	ba->Append(pByteArray->Data(),pByteArray->Size());
	i = pByteArray->Size();
	TransferThread.UnLock();
	return i;
}

/** */
void CTransfer::InitTime()
{
	TransferThread.Lock();
	starttime = time(0);
	ratetime  = starttime;
	currtime  = starttime;
	TransferThread.UnLock();
}

/** */
ulonglong CTransfer::GetTransferrate()
{
	return m_nCurrRate;
/*
	TransferThread.Lock();

	ulonglong r,t;
	
	if ( (starttime == 0) || (lTransfered == 0) || (m_bIdle == TRUE) )
	{
		TransferThread.UnLock();
		return 0;
	}

	if ( (eMode == estTRANSFERDOWNLOAD) ||
	     (eMode == estTRANSFERUPLOAD)  )
	{
		currtime = time(0);
	}

	r = lTransfered;
	t = currtime-starttime;

	if ( t > 0 )
	{
		r /= t;
	}

	TransferThread.UnLock();

	return r;*/
}

/** */
ulonglong CTransfer::GetBytesForTransferrate( ulonglong rate )
{
	TransferThread.Lock();

	ulonglong r = 0, t;

	if ( starttime == 0 )
	{
		TransferThread.UnLock();
		return 0;
	}

	if ( (eMode == estTRANSFERDOWNLOAD) ||
	     (eMode == estTRANSFERUPLOAD) )
	{
		currtime = time(0);
	}

	t = currtime-ratetime;

	//printf("RATE: %llu %llu %llu\n",t,m_nCurrRate,rate);

	r = m_nCurrRate;

	if ( t == 0 )
	{
		if ( (lTransfered - ratetransfered) > 0 )
			r = (r + lTransfered - ratetransfered)/2;
	}

	if ( r < rate )
		r = rate-r;
	else if ( (t == 0) || (r > rate) )
		r = 0;

	TransferThread.UnLock();

	return r;
}

/** */
void CTransfer::ConnectionState( eConnectionState state )
{
	CMessageConnectionState *Object;

	Object = new CMessageConnectionState();

	Object->m_eType = DC_MESSAGE_CONNECTION_STATE;

	Object->eState   = state;
	Object->sMessage = GetSocketError();

	if ( state == estCONNECTED )
	{
		if ( GetSocketMode() == esmSOCKET )
		{
			eMode = estTRANSFERHANDSHAKE;

			SendMyNick( sNick );
		}
		else
		{
			printf("change to ssl mode success\n");

			CallBack_SendObject(new CMessageKey());

		}
	}
	else if ( state == estDISCONNECTED )
	{
		eMode = estNONE;
	}

	CallBack_SendObject(Object);
}

/** */
int CTransfer::StartDownload( CString dstfile, ulonglong startposition, ulonglong currentposition, ulonglong endposition, ulonglong size, ulonglong chunksize, CString srcfile )
{
	if ( (srcfile == "" ) && (eMedium != eltBUFFER) )
	{
		printf("ctransfer: wrong mode %d\n",eMedium);
		return -1;
	}

	if ( m_bIdle == FALSE )
	{
		printf("ctransfer: other transfer is running\n");
		return -1;
	}

	if ( eMode == estTRANSFERUPLOAD )
	{
		printf("ctransfer: wrong transfer mode\n");
		return -1;
	}

	SetMode(estTRANSFERDOWNLOAD);
	SetStartPosition(startposition);
	SetCurrentPosition(currentposition);
	SetEndPosition(endposition);
	SetLength(size);
	SetDstFilename(dstfile);
	SetSrcFilename(srcfile);
	lTransfered = 0;
	lChunkSize = chunksize;
	ratetransfered = 0;
	m_nCurrRate = 0;
	m_nCurrRateCount = 0;
	memset( m_nCurrRateCache,0, sizeof(m_nCurrRateCache) );
	m_nCurrRateIndex = 0;

	/** set starttime */
	InitTime();

#ifdef HAVE_LIBBZ2
	if ( (GetDstFilename() == DC_USER_FILELIST_HE3) && (m_MessageSupports.m_bBZList == TRUE) )
	{
		SendGet( DC_USER_FILELIST_BZ, GetCurrentPosition()+1 );
	}
	else
#endif
	{
		if ( (m_MessageSupports.m_bChunk == FALSE) || (chunksize == 0) )
		{
			SendGet( GetDstFilename(), GetCurrentPosition()+1 );
		}
		else
		{
			SendGet( GetDstFilename(), GetCurrentPosition()+1, chunksize );
		}
	}
	
	return 0;
}

/** */
int CTransfer::StartUpload( CString dstfile, ulonglong length, ulonglong pos, ulonglong chunksize, CString srcfile )
{
	if ( (srcfile == "" ) && (eMedium != eltBUFFER) )
	{
		printf("ctransfer: wrong mode %d\n",eMedium);
		return -1;
	}

	if ( m_bIdle == FALSE )
	{
		printf("ctransfer: other transfer is running\n");
		return -1;
	}

	if ( eMode == estTRANSFERDOWNLOAD )
	{
		printf("ctransfer: wrong transfer mode\n");
		return -1;
	}

	SetMode(estTRANSFERUPLOAD);

	SetStartPosition(pos);
	SetCurrentPosition(pos);
	SetEndPosition(length);
	SetLength(length);
	SetDstFilename(dstfile);
	SetSrcFilename(srcfile);
	lTransfered = 0;
	lChunkSize = chunksize;
	ratetransfered = 0;
	m_nCurrRate = 0;
	m_nCurrRateCount = 0;
	memset( m_nCurrRateCache,0, sizeof(m_nCurrRateCache) );
	m_nCurrRateIndex = 0;

	/** set starttime */
	InitTime();

	SendFileLength(length);

	return 0;
}

/** */
int CTransfer::HandleMessage( char * c, int )
{
	int p=0;
	eDCMessage type;
	CObject * Object;
	CDir dir;

	CMessageHandler MessageHandler;

	CString s;

	s = c;

	while( (type=MessageHandler.Parse(&s,p,Object=0)) != DC_MESSAGE_PARSE_ERROR )
	{
		printf("transfer: handle message: %d\n",type);

		if ( !Object )
		{
			continue;
		}

		switch (type)
		{
			case DC_MESSAGE_FILELENGTH:
			case DC_MESSAGE_LISTLEN:
			{
				bool b = FALSE;
				CMessageFileLength * msg = (CMessageFileLength*)Object;

				if ( eMode != estTRANSFERDOWNLOAD || (m_bIdle == FALSE) )
				{
					printf("Warning: not in downloadmode/idle FILELENGTH/LISTLEN\n");

					delete Object;
					Object = 0;
				}
				else
				{
					lTransfered = 0;
					ratetransfered = 0;

					// update endposition
					if ( (lEndPosition == lLength) || (lEndPosition == 0) )
						lEndPosition = msg->lFileLength;

					lLength = msg->lFileLength;

					if ( eMedium == eltFILE )
					{
						if ( fd != -1 )
						{
#ifdef WIN32
							_close(fd);
#else
							close(fd);
#endif
						}
#ifdef WIN32
						fd = _open(sSrcFilename.Data(), _O_BINARY|_O_CREAT|_O_WRONLY, _S_IREAD|_S_IWRITE);

						if ( fd == -1 )
						{
							CallBack_SendError(strerror(errno));
							perror("file open");
						}
#else
						dir.SetPath("");

						if ( dir.IsFile(sSrcFilename) == FALSE )
						{
							fd = open(sSrcFilename.Data(), O_CREAT|O_WRONLY, S_IRWXU|S_IRGRP|S_IROTH);

							if ( fd == -1 )
							{
								CallBack_SendError(strerror(errno));
								perror("file open");
							}
						}
						else
						{
							fd = open(sSrcFilename.Data(),O_WRONLY);

							if ( fd == -1 )
							{
								CallBack_SendError(strerror(errno));
								perror("file open");
							}
						}
#endif
						if ( fd != -1 )
						{
							// jump to the startposition
#ifdef WIN32
							if ( _lseek(fd,(long)lCurrentPosition,SEEK_SET) != (long int)lCurrentPosition )
#else
							if ( lseek(fd,(long)lCurrentPosition,SEEK_SET) != (long int)lCurrentPosition )
#endif
							{
								CallBack_SendError(strerror(errno));
								perror("file seek");
#ifdef WIN32
								_close(fd);
#else
								close(fd);
#endif
								fd = -1;
							}
							else
							{
								b = TRUE;
							}
						}
					}
					else
					{
						pByteArray->SetSize(0);
						b = TRUE;
					}

					// start transfer
					if ( b == TRUE )
					{
						m_bIdle = FALSE;

						SendSend();
					}
				}

				break;
			}

			case DC_MESSAGE_SEND:
			{
				if ( (eMode != estTRANSFERUPLOAD) || (m_bIdle == FALSE) )
				{
					printf("Warning: not in uploadmode/idle SEND [%d/%d]\n",eMode,m_bIdle);

					delete Object;
					Object = 0;
				}
				else
				{
					lTransfered = 0;
					ratetransfered = 0;

					// init filehandle
					if ( eMedium == eltFILE )
					{
						lBufferPos        = SEND_FILE_BUFFER_SIZE;
						m_lReadBufferSize = SEND_FILE_BUFFER_SIZE;

						pByteArray->SetSize(SEND_FILE_BUFFER_SIZE);

						if ( fd != -1 )
						{
#ifdef WIN32
							_close(fd);
#else
							close(fd);
#endif
						}
#ifdef WIN32
						fd = _open(sSrcFilename.Data(),_O_BINARY|_O_RDONLY);
#else
						fd = open(sSrcFilename.Data(),O_RDONLY);
#endif
						if ( fd == -1 )
						{
							CallBack_SendError(strerror(errno));
							perror("file open");
						}
						else
						{
#ifdef WIN32
							if ( _lseek(fd,(long)lCurrentPosition,SEEK_SET) != (long int)lCurrentPosition )
#else
							if ( lseek(fd,(long)lCurrentPosition,SEEK_SET) != (long int)lCurrentPosition )
#endif
							{
								CallBack_SendError(strerror(errno));
								perror("file seek");
#ifdef WIN32
								_close(fd);
#else
								close(fd);
#endif
								fd = -1;
							}
						}

						if ( fd == -1 )
						{
							Disconnect(TRUE);
						}
						else
						{
							// start transfer
							m_bIdle = FALSE;
						}
					}
					else
					{
						// start transfer
						m_bIdle = FALSE;
					}

					printf("start upload ...'%s'\n",sSrcFilename.Data());
				}

				break;
			}

			case DC_MESSAGE_ERROR:
			{
				break;
			}

			case DC_MESSAGE_DIRECTION:
			{
				CMessageDirection * msg = (CMessageDirection*)Object;

				if ( eDstDirection != edNONE )
				{
					// remote will change the direction ... hoho ;-)
					if ( eDstDirection != msg->Direction )
					{
						Disconnect(TRUE);
						printf("Warning: remote will change his direction (hack) ...\n");
						delete Object;
						Object = 0;
					}
				}
				else
				{
					eDstDirection = msg->Direction;
					m_nDstLevel   = msg->iLevel;
				}

				break;
			}

			case DC_MESSAGE_MYNICK:
			{
				CMessageMyNick * msg = (CMessageMyNick*)Object;

				sDstNick = msg->sMyNick;

				m_nSrcLevel = rand()%0x7FFF;

				break;
			}

			case DC_MESSAGE_LOCK:
			{
				CString s;
				CEncrypt Encrypt;
				CMessageLock * msg = (CMessageLock*)Object;

				if ( eSrcDirection != edNONE )
				{
					// send supports command
					if ( msg->bExtProtocol == TRUE )
					{
						s = "";
#ifdef HAVE_LIBBZ2
						s += "BZList ";
#endif
#ifdef HAVE_SSL
						if ( (CConfig::Instance()->GetTransferCert() != "") &&
						     (CConfig::Instance()->GetTransferKey() != "") )
						{
							s += "SSL ";
						}
#endif
						// add chunk support
						s += "CHUNK ";

						if ( s != "" )
						{
							SendSupports(s);
						}
					}

					SendDirection( eSrcDirection, m_nSrcLevel );

					s = "";
					Encrypt.Encrypt(msg->sData,s);

					SendKey(s);
				}
				else
				{
					Disconnect(TRUE);
				}

				break;
			}

			case DC_MESSAGE_SUPPORTS:
			{
				CMessageSupports * msg = (CMessageSupports*)Object;

				m_MessageSupports = *msg;

				break;
			}

			case DC_MESSAGE_KEY:
			{
#ifdef HAVE_SSL
				// start ssl socket
				if ( (m_MessageSupports.m_bSSL == TRUE) &&
				     (CConfig::Instance()->GetTransferCert() != "") &&
				     (CConfig::Instance()->GetTransferKey() != "") )
				{
					bool b;
					if ( m_bListener == TRUE )
						b = ChangeSocketMode(esmSSLSERVER, CConfig::Instance()->GetTransferCert(), CConfig::Instance()->GetTransferKey() );
					else
						b = ChangeSocketMode(esmSSLCLIENT);
					printf("change to ssl mode %d\n",b);
					delete Object;
					Object = 0;
				}
#endif

				break;
			}

			case DC_MESSAGE_MAXEDOUT:
			{
				break;
			}

			case DC_MESSAGE_CANCEL:
			{
				if ( (eMode == estTRANSFERUPLOAD) && (m_bIdle == FALSE) )
				{
					if ( eMedium == eltFILE )
					{
#ifdef WIN32
						_close(fd);
#else
						close(fd);
#endif
						fd = -1;
					}

					// free buffer
					pByteArray->SetSize(0);

					// stop transfer
					m_bIdle = TRUE;
				}
				else
				{
					printf("Warning: Cancel msg not in uploadmode/idle SEND [%d/%d]\n",eMode,m_bIdle);

					delete Object;
					Object = 0;
				}

				break;
			}

			default:
			{
				//printf("ctransfer: unknown message %d\n",type);
			}
		}

		if (Object)
		{
			((CDCMessage*)Object)->m_eType = type;

			CallBack_SendObject(Object);
		}
	}

	return 0;
}

/** */
bool CTransfer::GetEncrypted()
{
	bool res = FALSE;

#ifdef HAVE_SSL
	if ( (GetSocketMode() == esmSSLCLIENT) ||
	     (GetSocketMode() == esmSSLSERVER) )
	{
		res = TRUE;
	}
#endif

	return res;
}

/** */
void CTransfer::DataSend()
{
	int l = 0;
	ulonglong len;
	ulonglong sendbyte;
	ulonglong freebuf;
	CMessageTransfer *Object;

	if ( (eMode == estTRANSFERUPLOAD) && (m_bIdle == FALSE) )
	{
		sendbyte = 0;

		// check send rate
		if ( lRate != 0 )
		{
			sendbyte = GetBytesForTransferrate(lRate);

			if ( sendbyte == 0 )
			{
				return;
			}
		}

		freebuf = 0xFFFF; //GetFreeSendBufferSize();

		if ( freebuf == 0 )
		{
			return;
		}

		if ( eMedium == eltFILE )
		{
			if ( (m_MessageSupports.m_bChunk == FALSE) || (lChunkSize == 0) )
			{
				len = lLength - lCurrentPosition;
			}
			else
			{
				len = lStartPosition+lChunkSize - lCurrentPosition;
			}

			//printf("%llu %llu %llu %llu\n",len,lStartPosition,lChunkSize,lCurrentPosition);

			// load data from filehandle and send ...
			if ( lBufferPos == m_lReadBufferSize )
			{
				if ( len > SEND_FILE_BUFFER_SIZE )
					len = SEND_FILE_BUFFER_SIZE;

#ifdef WIN32
				m_lReadBufferSize = _read( fd, pByteArray->Data(), (unsigned int)len );
#else
				m_lReadBufferSize = read( fd, pByteArray->Data(), (unsigned int)len );
#endif
				if ( m_lReadBufferSize == -1 )
				{
					CallBack_SendError(strerror(errno));
					perror("CTransfer::DataSend() read error ! ");
					return;
				}
				else if ( m_lReadBufferSize == 0 )
				{
					perror("CTransfer::DataSend() no data read ! ");
					Disconnect(TRUE);
					return;
				}
				else if ( len > (ulonglong)m_lReadBufferSize )
				{
					perror("CTransfer::DataSend() wrong length calculation ! ");
					len = m_lReadBufferSize;
				}

				lBufferPos = 0;
			}

			// limit to transferrate
			if ( sendbyte != 0 )
			{
				if ( len > sendbyte )
				{
					len = sendbyte;
				}
			}

			// limit to free socket sendbuffer
			if ( len > freebuf )
			{
				len = freebuf;
			}

			// limit to sendbuffer
			if ( len > (ulonglong)(SEND_FILE_BUFFER_SIZE-lBufferPos) )
			{
				len = SEND_FILE_BUFFER_SIZE-lBufferPos;
			}

			l = Write( pByteArray->Data()+lBufferPos, (long)len, TRUE );
			//printf("WRITE %llu %llu %llu -> %d\n",lCurrentPosition,lLength,len,l);
			if ( l > 0 )
			{
				lTransfered += l;
				lCurrentPosition += l;
				lBufferPos += l;

				// add traffic control
				CSocket::m_Traffic.AddTraffic(ettDATATX,len);
			}
			else if ( l == -1 )
			{
				perror("CTransfer::DataSend() write error !\n");
			}
			else if ( l == 0 )
			{
				return;
			}
		}
		else if ( eMedium == eltBUFFER )
		{
			// send data from byte array ...
			// todo calc speed etc.
			len = lLength - lCurrentPosition;

			// limit transferrate
			if ( sendbyte != 0 )
			{
				if ( len > sendbyte )
				{
					len = sendbyte;
				}
			}

			// limit to free socket sendbuffer
			if ( len > freebuf )
			{
				len = freebuf;
			}

			l = Write( pByteArray->Data()+lTransfered, (long)len, TRUE );

			if ( l > 0 )
			{
				lTransfered += l;
				lCurrentPosition += l;

				// add traffic control
				CSocket::m_Traffic.AddTraffic(ettDATATX,len);
			}
			else if ( l == -1 )
			{
				perror("CTransfer::DataSend() write error !\n");
			}
			else if ( l == 0 )
			{
				return;
			}
		}
		else
		{
			// error ... wrong medium
		}

		if ( (lLength == lCurrentPosition) || ((m_MessageSupports.m_bChunk == TRUE) && ((lStartPosition+lChunkSize) == lCurrentPosition)) )
		{
			printf("end found\n");
			if ( eMedium == eltFILE )
			{
#ifdef WIN32
				_close(fd);
#else
				close(fd);
#endif
				fd = -1;
			}

			// free buffer
			pByteArray->SetSize(0);

			// stop transfer
			m_bIdle = TRUE;
		}

		/** update currtime */
		currtime = time(0);

		Object = new CMessageTransfer();
		Object->m_eType = DC_MESSAGE_TRANSFER;
		Object->iMode = 1;
		Object->lTransfered = lCurrentPosition;
		Object->lLength = lLength;

		CallBack_SendObject(Object);
	}
}

/** */
void CTransfer::DataTimeout()
{
	printf("CTransfer: data timeout ...\n");

	if ( (eMode == estTRANSFERDOWNLOAD) && (m_bIdle == FALSE) )
	{
		// send send ...
		SendSend();
	}
	else if ( eMode == estTRANSFERHANDSHAKE )
	{
		// handshake timeout
		Disconnect(TRUE);
	}
	else if ( m_bIdle == TRUE )
	{
		// send string ...
		SendString("|");
	}
}

/** */
void CTransfer::Notify()
{
	ulonglong r,t;

	TransferThread.Lock();

	if ( (time(0)-ratetime) < 1 )
	{
		TransferThread.UnLock();
		return;
	}

	if ( (starttime == 0) || (lTransfered == 0) || (m_bIdle == TRUE) )
	{
		m_nCurrRate = 0;
		TransferThread.UnLock();
		return;
	}

	if ( (eMode == estTRANSFERDOWNLOAD) ||
	     (eMode == estTRANSFERUPLOAD)  )
	{
		currtime = time(0);
	}

	r = lTransfered-ratetransfered;
	t = currtime-ratetime;

	if ( t > 0 )
	{
		r /= t;
	}

	//printf("%llu %llu\n",lTransfered,r);

	ratetransfered = lTransfered;
	ratetime       = currtime;

	m_nCurrRateCache[m_nCurrRateIndex] = r;
	m_nCurrRateIndex++;
	if ( m_nCurrRateIndex == 10 )
		m_nCurrRateIndex = 0;
	if ( m_nCurrRateCount < 10 )
		m_nCurrRateCount++;
	t = 0;
	for(r=0;r<10;r++)
		t += m_nCurrRateCache[r];
	m_nCurrRate = t/m_nCurrRateCount;
	//printf("%llu %llu\n",t,m_nCurrRate);

	TransferThread.UnLock();
}

/** */
void CTransfer::DataAvailable( const char * buffer, int len )
{
	CMessageTransfer *Object;
	CString s;
	CString serr="";
	bool berr=FALSE;
	int i,i1;
	ulonglong l;

	if ( (eMode == estTRANSFERDOWNLOAD) && (m_bIdle == FALSE) )
	{
		// add traffic control
		CSocket::m_Traffic.AddTraffic(ettDATARX,len);

		// check endposition
		if ( (len+lCurrentPosition) > lEndPosition )
		{
			l = lEndPosition-lCurrentPosition;
		}
		else
		{
			l = len;
		}

		i1=0;

		if ( eMedium == eltFILE )
		{
			if ( (fd != -1) && (l>0) )
			{
				while(i1 != (long int)l)
				{
#ifdef WIN32
					i = _write(fd,buffer,(long)l);
#else
					i = write(fd,buffer,l);
#endif
					if (i!=-1)
					{
						i1+=i;
					}
					else
					{
						berr = TRUE;
						serr = strerror(errno);
						break;
					}
				}
			}
			else
			{
				// TODO: error message
			}
		}
		else if ( eMedium == eltBUFFER )
		{
			pByteArray->Append((const unsigned char*)buffer,(unsigned long)l);
			i1+=l;
		}
		else
		{
			// error ...
		}

		if ( i1 > 0 )
		{
			lTransfered+=i1;
			lCurrentPosition+=i1;
		}

//		printf("%ld %ld %ld %ld\n",lLength,lPos,lTransfered,lBufferPos);

		if ( lEndPosition == lCurrentPosition )
		{
			printf("CTransfer fileend found: %llu %llu %llu\n",lLength,lEndPosition,lCurrentPosition);

			// if endposition not filesize we must disconnect
			if ( (lEndPosition != lLength) && (m_MessageSupports.m_bChunk == FALSE) )
			{
				Disconnect(TRUE);
			}

			m_bIdle = TRUE;

			if ( eMedium == eltFILE )
			{
#ifdef WIN32
				_close(fd);
#else
				close(fd);
#endif

				fd = -1;
			}
			else if ( sDstFilename == DC_USER_FILELIST_HE3 )
			{
#ifdef HAVE_LIBBZ2
				// decompress filelist
				if ( m_MessageSupports.m_bBZList == FALSE )
#endif
				{
					CHE3 *he3=0;
					CString * s=0;

					he3 = new CHE3();
					s = he3->decode_he3_data(pByteArray);
					delete he3;
					pByteArray->SetSize(0);

					if ( s != 0 )
					{
						pByteArray->Append( (unsigned char*)s->Data(), s->Length() );
						delete s;
					}
				}
#ifdef HAVE_LIBBZ2
				else
				{
					CBZ * bz = 0;
					CByteArray out;

					bz = new CBZ();

					if ( bz->Decompress( pByteArray, &out ) == TRUE )
					{
						pByteArray->SetSize(0);
						pByteArray->Append( out.Data(), out.Size() );
					}
					else
					{
						printf("bz2 decompress failed\n");
						pByteArray->SetSize(0);
					}

					delete bz;
				}
#endif
			}
		}

		if ( i1 > 0 )
		{
			Object = new CMessageTransfer();
			Object->m_eType = DC_MESSAGE_TRANSFER;
			Object->iMode = 1;
			Object->lTransfered = lCurrentPosition;
			Object->lLength = lEndPosition;

			CallBack_SendObject(Object);
		}

		// send the error message to the appl
		if ( berr == TRUE )
		{
			CallBack_SendError(serr);

			// disconnect on error
			Disconnect(TRUE);
		}

		/** update currtime */
		currtime = time(0);

		return;
	}

	// add traffic control
	CSocket::m_Traffic.AddTraffic(ettCONTROLRX,len);

	i = i1 = 0;

	s = buffer;

/*
	// simple workaround for the wrong cancel message from nmdc
	if ( s == "$Canceled" )
		s += "|";
	else if ( s == "$Cancel" )
		s += "|";
*/
	s = sBuffer + s;

	// search '|'
	while((i=s.Find('|',i))>=0)
	{
		i++;
		i1 = i;
	}

	if ( i1 > 0 )
	{
		sBuffer = s.Mid(0,i1);
		if ( HandleMessage( (char*)sBuffer.Data(), sBuffer.Length() ) == -1 )
		{
		}
	}

	if (i1==s.Length())
	{
		sBuffer = "";
	}
	else
	{
		sBuffer = s.Mid( i1, s.Length()-i1 );
	}
}
